/**
 * Initializes the store
 *
 * @remarks
 *
 * **Note**: Ensure that the **ENTIRE** state is fully initialized in the createStore call, or you will faces issues with crashes.
 * If a prop must support undefined, define it as "foo: type | undefined" rather than "foo?: type".
 */
export default function createStore<T>(initialState: Required<T>, useReduxDevTools?: boolean): {
    /** Helper function to create "prebaked" update methods */
    action<T_1 extends unknown[]>(resolver: (store: Required<T>, ...args: T_1) => void | Partial<Required<T>> | Promise<void | Partial<Required<T>> | undefined> | undefined, affectedContexts?: (keyof T)[]): (...args: T_1) => Promise<{
        update(updatedValues: Partial<Required<T>>): void;
        get(): Readonly<Required<T>>;
        getValue(prop: keyof T): Required<T>[keyof T];
        setValue<K extends keyof T>(prop: K, value: Required<T>[K]): void;
        set(state: Required<Required<T>>): void;
        keys(): (keyof T)[];
    }>;
    /** Returns the entire global state */
    get(): Readonly<Required<T>>;
    /** Updates the pending status of the specified context */
    /**
     * Updates boolean status for **pending operations** in parts of the global state
     *
     * @remarks
     *
     * `pending` are to be used within an `action`.
     *
     * How to use `pending` within an `action`:
     *
     * ```ts
     * const foo = action(async () => {
     *  pending(["stateVar1", "stateVar2"], true);
     *
     *  const stateVar1 = await api("url1");
     *  const stateVar2 = await api("url2");
     *
     *  pending(["stateVar1", "stateVar2"], false);
     *  return { stateVar1, stateVar2 }
     * });
     * ```
     */
    pending<TContext extends keyof T>(contexts: TContext | TContext[], state: boolean): void;
    /** Sets up a subscription for a single global state context */
    subscribe<TContext_1 extends keyof T>(context: TContext_1, callback: (value: Required<T>[TContext_1]) => void): () => void;
    /**
     * Update one or more parts of the global state.
     *
     * @remarks
     *
     * `update` are to be used within an `action`. It allows for state updates before an action is finished.
     * The function receives an object that represents parts of the global state.
     *
     * How to use `update` within an `action`:
     *
     * ```ts
     * const foo = action(() => {
     *  ...
     *  update({ storeVar1: "foo", storeVar2: 1});
     *  ...
     * });
     * ```
     */
    update<TContext_2 extends keyof T>(value: Pick<Required<T>, TContext_2>): void;
    /**
     * Subscribe to boolean updates for **async operations** in parts of the global state
     *
     * @remarks
     *
     * How to use `usePending` within a functional component component:
     *
     * ```tsx
     * export const Comp = () => {
     *  const loading = usePending();
     *  ...
     *  if(loading.storeVar1)
     *      return "Loading..";
     *  else
     *      return <div>Content</div>
     * }
     * ```
     * See documentation for `pending()` for more details
     */
    usePending<T_2 extends keyof T>(...explicitContexts: T_2[]): Readonly<{ [K_1 in keyof T]: boolean; }>;
    /**
     * Subscribes to updates in the global state to re-render a component
     *
     * @remarks
     *
     * How to use `useSquawk` within a functional component:
     *
     * ```tsx
     * export const Comp = () => {
     *  const { storeVar1, storeVar2 } = useSquawk();
     *  ...
     *  return <div>{storevar1} - {storeVar2}</div>
     * }
     * ```
     */
    useSquawk<T_3 extends keyof T>(...explicitContexts: T_3[]): Required<T>;
};
